@@ -1,18 +1,22 @@
 package App::Termcast;
 BEGIN {
-  $App::Termcast::VERSION = '0.10';
+  $App::Termcast::AUTHORITY = 'cpan:DOY';
+}
+{
+  $App::Termcast::VERSION = '0.11';
 }
 use Moose;
 # ABSTRACT: broadcast your terminal sessions for remote viewing
 
 with 'MooseX::Getopt::Dashes';
 
-use IO::Pty::Easy;
+use IO::Select;
 use IO::Socket::INET;
-use Scope::Guard;
+use JSON;
+use Scalar::Util 'weaken';
+use Term::Filter::Callback;
 use Term::ReadKey;
 use Try::Tiny;
-use JSON;
 
 
 
@@ -65,14 +69,6 @@ has timeout => (
     documentation => "Timeout length for the connection to the termcast server",
 );
 
-has _got_winch => (
-    traits   => ['NoGetopt'],
-    is       => 'rw',
-    isa      => 'Bool',
-    default  => 0,
-    init_arg => undef,
-);
-
 
 has establishment_message => (
     traits     => ['NoGetopt'],
@@ -129,44 +125,61 @@ sub _build_socket {
                                         PeerPort => $self->port);
         if (!$socket) {
             Carp::carp "Couldn't connect to " . $self->host . ": $!";
+            ReadMode(0, $self->input)
+                if $self->_has_term && $self->_term->_raw_mode;
             sleep 5;
+            ReadMode(5, $self->input)
+                if $self->_has_term && $self->_term->_raw_mode;
             redo;
         }
     }
 
-    $socket->syswrite($self->establishment_message . $self->termsize_message);
+    syswrite $socket, $self->establishment_message . $self->termsize_message;
 
     # ensure the server accepted our connection info
-    # can't use _build_select_args, since that would cause recursion
     {
-        my ($rin, $ein, $rout, $eout) = ('') x 4;
-        vec($rin, fileno($socket), 1) = 1;
-        vec($ein, fileno($socket), 1) = 1;
-        my $res = select($rout = $rin, undef, $eout = $ein, undef);
-        redo if ($!{EAGAIN} || $!{EINTR}) && $res == -1;
-        if (vec($eout, fileno($socket), 1)) {
-            Carp::croak("Invalid password");
-        }
-        elsif (vec($rout, fileno($socket), 1)) {
-            my $buf;
-            $socket->recv($buf, 4096);
-            if (!defined $buf || length $buf == 0) {
+        my $select = IO::Select->new($socket);
+        my ($r, undef, $e) = IO::Select->select(
+            $select, undef, $select,
+        );
+
+        for my $fh (@$e) {
+            if ($fh == $socket) {
+                ReadMode(0, $self->input)
+                    if $self->_has_term && $self->_term->_raw_mode;
                 Carp::croak("Invalid password");
             }
-            elsif ($buf ne ('hello, ' . $self->user . "\n")) {
-                Carp::carp("Unknown login response from server: $buf");
+        }
+        for my $fh (@$r) {
+            if ($fh == $socket) {
+                my $buf;
+                $socket->recv($buf, 4096);
+                if (!defined $buf || length $buf == 0) {
+                    ReadMode(0, $self->input)
+                        if $self->_has_term && $self->_term->_raw_mode;
+                    Carp::croak("Invalid password");
+                }
+                elsif ($buf ne ('hello, ' . $self->user . "\n")) {
+                    ReadMode(0, $self->input)
+                        if $self->_has_term && $self->_term->_raw_mode;
+                    Carp::carp("Unknown login response from server: $buf");
+                    ReadMode(5, $self->input)
+                        if $self->_has_term && $self->_term->_raw_mode;
+                }
             }
         }
     }
 
-    ReadMode 5 if $self->_raw_mode;
+    ReadMode(5, $self->input)
+        if $self->_has_term && $self->_term->_raw_mode;
     return $socket;
 }
 
 before clear_socket => sub {
     my $self = shift;
     Carp::carp("Lost connection to server ($!), reconnecting...");
-    ReadMode 0 if $self->_raw_mode;
+    ReadMode(0, $self->input)
+        if $self->_has_term && $self->_term->_raw_mode;
 };
 
 sub _new_socket {
@@ -175,35 +188,6 @@ sub _new_socket {
     $self->socket;
 }
 
-has pty => (
-    traits     => ['NoGetopt'],
-    is         => 'rw',
-    isa        => 'IO::Pty::Easy',
-    lazy_build => 1,
-    init_arg   => undef,
-);
-
-sub _build_pty {
-    IO::Pty::Easy->new(raw => 0);
-}
-
-has _raw_mode => (
-    traits  => ['NoGetopt'],
-    is      => 'rw',
-    isa     => 'Bool',
-    default => 0,
-    trigger => sub {
-        my $self = shift;
-        my ($val) = @_;
-        if ($val) {
-            ReadMode 5;
-        }
-        else {
-            ReadMode 0;
-        }
-    },
-);
-
 has _needs_termsize_update => (
     traits  => ['NoGetopt'],
     is      => 'rw',
@@ -211,153 +195,101 @@ has _needs_termsize_update => (
     default => 0,
 );
 
-sub _build_select_args {
-    my $self = shift;
-    my @for = @_ ? @_ : (qw(socket pty input));
-    my %for = map { $_ => 1 } @for;
+has _term => (
+    is        => 'ro',
+    does      => 'Term::Filter',
+    lazy      => 1,
+    predicate => '_has_term',
+    default   => sub {
+        my $_self = shift;
+        weaken(my $self = $_self);
+        # XXX using ::Callback for now because we need to be able to
+        # instantiate App::Termcast objects without initializing the terminal
+        # (in case of just calling write_to_termcast). This should
+        # eventually be deprecated in favor of moving the termcast interaction
+        # code out to an App::Termcast::Writer module or something, and
+        # this module should be a simple wrapper that combines that module
+        # with Term::Filter.
+        Term::Filter::Callback->new(
+            callbacks => {
+                setup => sub {
+                    my ($term) = @_;
+                    $term->add_input_handle($self->socket);
+                },
+                winch => sub {
+                    # for the sake of sending a clear to the client anyway
+                    syswrite $self->output, "\e[H\e[2J";
+                    $self->_needs_termsize_update(1);
+                },
+                read_error => sub {
+                    my ($term, $fh) = @_;
+                    if ($fh == $self->socket) {
+                        $self->_new_socket;
+                    }
+                },
+                read => sub {
+                    my ($term, $fh) = @_;
+                    if ($fh == $self->socket) {
+                        my $got = $term->_read_from_handle(
+                            $self->socket, "socket"
+                        );
+                        $self->_new_socket unless defined $got;
 
-    my ($rin, $win, $ein) = ('', '', '');
-    if ($for{socket}) {
-        my $sockfd = fileno($self->socket);
-        vec($rin, $sockfd, 1) = 1;
-        vec($win, $sockfd, 1) = 1;
-        vec($ein, $sockfd, 1) = 1;
-    }
-    if ($for{pty}) {
-        my $ptyfd  = fileno($self->pty);
-        vec($rin, $ptyfd,  1) = 1;
-    }
-    if ($for{input}) {
-        my $infd   = fileno(STDIN);
-        vec($rin, $infd   ,1) = 1;
-    }
-
-    return ($rin, $win, $ein);
-}
-
-sub _socket_ready {
-    my $self = shift;
-    my ($vec) = @_;
-    vec($vec, fileno($self->socket), 1);
-}
-
-sub _pty_ready {
-    my $self = shift;
-    my ($vec) = @_;
-    vec($vec, fileno($self->pty), 1);
-}
-
-sub _in_ready {
-    my $self = shift;
-    my ($vec) = @_;
-    vec($vec, fileno(STDIN), 1);
-}
+                        if ($self->bell_on_watcher) {
+                            # something better to do here?
+                            syswrite $self->output, "\a";
+                        }
+                    }
+                },
+                munge_output => sub {
+                    my ($term, $buf) = @_;
+                    $self->write_to_termcast($buf);
+                    $buf;
+                },
+            },
+        );
+    },
+    handles => [ 'run', 'input', 'output' ],
+);
 
 
 sub write_to_termcast {
     my $self = shift;
     my ($buf) = @_;
 
-    my ($rin, $win, $ein) = $self->_build_select_args('socket');
-    my ($rout, $wout, $eout);
-    my $ready = select(undef, $wout = $win, $eout = $ein, $self->timeout);
-    if (!$ready || $self->_socket_ready($eout)) {
-        $self->clear_socket;
-        return $self->write_to_termcast(@_);
+    my $socket = $self->socket;
+    my $select = IO::Select->new($socket);
+
+    my (undef, $w, $e) = IO::Select->select(
+        undef, $select, $select, $self->timeout,
+    );
+
+    my $err;
+
+    for my $fh (@$e) {
+        if ($fh == $socket) {
+            $err = 1;
+        }
     }
 
-    if ($self->_needs_termsize_update) {
-        $buf = $self->termsize_message . $buf;
-        $self->_needs_termsize_update(0);
+    if (!$err) {
+        for my $fh (@$w) {
+            if ($fh == $socket) {
+                if ($self->_needs_termsize_update) {
+                    $buf = $self->termsize_message . $buf;
+                    $self->_needs_termsize_update(0);
+                }
+
+                return $self->socket->syswrite($buf);
+            }
+        }
     }
 
-    $self->socket->syswrite($buf);
+    $self->clear_socket;
+    return $self->write_to_termcast(@_);
 }
 
 
-sub run {
-    my $self = shift;
-    my @cmd = @_;
-
-    $self->socket;
-
-    $self->_raw_mode(1);
-    my $guard = Scope::Guard->new(sub { $self->_raw_mode(0) });
-
-    $self->pty->spawn(@cmd) || die "Couldn't spawn @cmd: $!";
-
-    local $SIG{WINCH} = sub {
-        $self->_got_winch(1);
-        $self->pty->slave->clone_winsize_from(\*STDIN);
-
-        $self->pty->kill('WINCH', 1);
-
-        syswrite STDOUT, "\e[H\e[2J"; # for the sake of sending a
-                                      # clear to the client anyway
-
-        $self->_needs_termsize_update(1);
-    };
-
-    while (1) {
-        my ($rin, $win, $ein) = $self->_build_select_args;
-        my ($rout, $wout, $eout);
-        my $select_res = select($rout = $rin, undef, $eout = $ein, undef);
-        my $again = $!{EAGAIN} || $!{EINTR};
-
-        if (($select_res == -1 && $again) || $self->_got_winch) {
-            $self->_got_winch(0);
-            redo;
-        }
-
-        if ($self->_socket_ready($eout)) {
-            $self->_new_socket;
-        }
-
-        if ($self->_in_ready($rout)) {
-            my $buf;
-            sysread STDIN, $buf, 4096;
-            if (!defined $buf || length $buf == 0) {
-                Carp::croak("Error reading from stdin: $!")
-                    unless defined $buf;
-                last;
-            }
-
-            $self->pty->write($buf);
-        }
-
-        if ($self->_pty_ready($rout)) {
-            my $buf = $self->pty->read(0);
-            if (!defined $buf || length $buf == 0) {
-                Carp::croak("Error reading from pty: $!")
-                    unless defined $buf;
-                last;
-            }
-
-            syswrite STDOUT, $buf;
-
-            $self->write_to_termcast($buf);
-        }
-
-        if ($self->_socket_ready($rout)) {
-            my $buf;
-            $self->socket->recv($buf, 4096);
-            if (!defined $buf || length $buf == 0) {
-                if (defined $buf) {
-                    $self->_new_socket;
-                }
-                else {
-                    Carp::croak("Error reading from socket: $!");
-                }
-            }
-
-            if ($self->bell_on_watcher) {
-                # something better to do here?
-                syswrite STDOUT, "\a";
-            }
-        }
-    }
-}
-
 __PACKAGE__->meta->make_immutable;
 no Moose;
 
@@ -373,7 +305,7 @@ App::Termcast - broadcast your terminal sessions for remote viewing
 
 =head1 VERSION
 
-version 0.10
+version 0.11
 
 =head1 SYNOPSIS
 
@@ -436,9 +368,9 @@ Sends C<$BUF> to the termcast server.
 
 =head2 run @ARGV
 
-Runs the given command in the local terminal as though via C<exec>, but streams
-all output from that command to the termcast server. The command may be an
-interactive program (in fact, this is the most useful case).
+Runs the given command in the local terminal as though via C<system>, but
+streams all output from that command to the termcast server. The command may be
+an interactive program (in fact, this is the most useful case).
 
 =head1 TODO
 
@@ -454,16 +386,8 @@ L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=App-Termcast>.
 
 =head1 SEE ALSO
 
-Please see those modules/websites for more information related to this module.
-
-=over 4
-
-=item *
-
 L<http://termcast.org/>
 
-=back
-
 =head1 SUPPORT
 
 You can find this documentation for this module with the perldoc command.
@@ -498,7 +422,7 @@ Jesse Luehrs <doy at tozt dot net>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2011 by Jesse Luehrs.
+This software is copyright (c) 2012 by Jesse Luehrs.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
